﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Runtime.InteropServices;
using System.Windows.Threading;
using System.Threading;

namespace XEAS
{
	/// <summary>
	/// Interaction logic for Window1.xaml
	/// </summary>
	public partial class Window1 : Window
	{
		//=============================================================================================================
		// Data block returned on 'poll'
		//=============================================================================================================
		// this is a hack of the C++ structure used in the WINEAS based on the C in the box
		// it's 'unsafe' because they are not managed variables. Hence the use of 'fixed' to
		// qualify the arrays and because I really wanred a union with packing set to 1 I am
		// forced to use the FieldOffset trick but that means I need to set it for each and
		// every variable in the set.
		[StructLayout(LayoutKind.Explicit, Size=298)]
		unsafe struct EAS {
			[FieldOffset(0)]
			public byte RAM_START;				// 'S'
			// debug hold via SW2
			[FieldOffset(1)]
			public byte bHold;
			// Time values
			[FieldOffset(2)]
			public ushort mSecs;				// 0-999
			[FieldOffset(4)]
			public ushort secs;					// 0 to 18 hours
			// A2D values for suspension
			// Readings are taken every quarter second
			// ports are ( 0,1,2,4 )=( RF,LF,RR,LR )
			[FieldOffset(6)]
			public byte adN;					// current A2D channel index 0-3
			[FieldOffset(7)]
			public byte adIndex;				// index to averages (points to most recent)
			[FieldOffset(8)]
			public fixed ushort adRaw[4];		// current 'raw' values 0-1023 = 0-5V
			// #define AD_MAX	900				// out of range tests (about 930 when n/c)
			// #define AD_MIN	300
			[FieldOffset(16)]
			public fixed ushort adFinal[4];		// processed height in mm
			// Then we need an averaged version for when we are on the move
			// I choose to do readings every quarter second and keep 2 seconds worth
			// #define NSLOTS	20				// 4 per second for 2 seconds
			[FieldOffset(24)]
			public fixed ushort hLag[80];		// [20][4];	// buffer
			// These are then processed and normalised to give the final version
			[FieldOffset(184)]
			public fixed ushort _height[4];		// translated values in mm
			// #define hRF	x.height[0]			// and a few shorthands
			// #define hLF x.height[1]
			// #define hRR x.height[2]
			// #define hLR x.height[3]
			[FieldOffset(192)]
			public byte nLag;					// slot pointer

			// Pulse values
			[FieldOffset(193)]
			public byte rpmTicks;				// RPM service counter
			[FieldOffset(194)]
			public ushort prevRPM;				// previous counter
			[FieldOffset(196)]
			public ushort rawRPM;				// RPM counts per pulse
			[FieldOffset(198)]
			public ushort RPM;					// processed RPM
			[FieldOffset(200)]
			public byte mphTicks;				// MPH counts per second counter
			[FieldOffset(201)]
			public ushort rawMPH;				// MPH counts per second
			[FieldOffset(203)]
			public ushort MPH;					// processed MPH

			// Compressor operation
			[FieldOffset(205)]
			public byte compON;					// we are compressing
			[FieldOffset(206)]
			public ushort compTimer;			// pump too long = leak
			// void compressorTick();			// Protect forward references
			// #define COMP_LIMIT	600			// ten minutes at 10 counts per second

			// Command system
			[FieldOffset(208)]
			public fixed byte usbCmd[19];		// command line
			[FieldOffset(227)]
			public byte usbNcmd;				// number of bytes in the command line

			// Ride height switch system
			// These are the 'official' values.
			// If xBUGS or xINHIBIT are set I may well do my own thing
			[FieldOffset(228)]
			public byte rhLights;				// 1=HIGH 2=NORM 4=MWAY 8=ACCESS
			[FieldOffset(229)]
			public byte rhFlash;				// as above but flashing
			[FieldOffset(230)]
			public byte rhAlt;					// alternating pass flag to do the flash
			[FieldOffset(231)]
			public byte rhWork;					// working values

			// 'high speed' logic
			[FieldOffset(232)]
			public byte bHighSpeed;				// logic flag
			[FieldOffset(233)]
			public byte nHighSpeed;				// seconds counter

			// Mode selection and transition
			// We have officially three selectable modes, I add two more
			// enum UMODE { M_BUMPSTOPS, M_ACCESS, M_NORMAL, M_HIGH, M_VERYHIGH };
			[FieldOffset(234)]
			public byte uMode;					// M_NORMAL desired mode
			// However this translates into a more complex mode when you add the rest
			// enum SMODE { S_BUMPSTOPS, S_ACCESS, S_LOW, S_NORMAL, S_HIGH, S_VERYHIGH, S_UNKNOWN };
			[FieldOffset(235)]
			public byte sMode;					// S_UNKNOWN  current mode
			[FieldOffset(236)]
			public byte tMode;					// S_NORMAL   target mode
			[FieldOffset(237)]
			public byte tHeight;				// target height
			// and we need a set of proposed heights for each mode
			[FieldOffset(238)]
			public fixed byte mHeight[6];
			//	0,		// Bumpstops
			//	10,		// Access			-2.5"  (-63mm)
			//	48,		// Low Profile		-1.0"  (-25mm)
			//	73,		// Standard Profile	 0"	   ( 0mm)
			//	112,	// High Profile		+1.5"  (+38mm)
			//	130		// Extended Profile	+2.25" (+57mm)

			// enum { M_HOLD,											// no planned move
			//		  M_URF, M_ULF, M_URR, M_ULR, M_UF, M_UR,			// up moves
			//		  M_DRF, M_DLF, M_DRR, M_DLR, M_DF, M_DR, M_D,		// down moves
			//		  M_SETTLE=0x80 };									// apres move dwell
			[FieldOffset(244)]
			public byte mState;					// what we are doing
			[FieldOffset(245)]
			public byte tState;					// state working timer (tenths of a second)
			[FieldOffset(246)]
			public byte step;						// step change
			// To access the extra modes we need some button hold counters...
			[FieldOffset(247)]
			public byte upHold;
			[FieldOffset(248)]
			public byte downHold;

			// eeWrite system
			[FieldOffset(249)]
			public ushort eeAddr;					// counter for bytes written

			// The bug system
			// Any bit set here is a catastrophic bug and we have to stop running things.
			// However notice that in my world switching off and on again gives things a
			// second chance

			[FieldOffset(251)]
			public ushort bugs;
			// these are the bug bits
			// #define bugOVERHEAT		0		// overheat flag set
			// #define bugOVERRUN		1		// compressor ran for too long
			// #define bugSTRUTHI		2		// over volts for 4 struts
			// #define bugSTRUTLO		6		// under volts for 4 struts
			// #define bugTOOHIGH		10		// one strut too high (debug)
			// #define bugFATAL			0	//0xFFFF	// for now all bugs are fatal

			// Constants block we read from the eerom
			[FieldOffset(253)]
			public fixed byte eeRom_errors[16];		// error flags
			[FieldOffset(269)]
			public fixed ushort eeRom_offsets[4];		// bump stop height offsets
			// #define EECOUNT	(sizeof(x.eeRom_errors)+sizeof(x.eeRom_offsets))
			// #define ERROR_COUNT sizeof(x.eeRom_errors)

			// For debugging I need some reported values in the dump
			[FieldOffset(277)]
			public ushort vDebug1;
			[FieldOffset(279)]
			public ushort vDebug2;			
			[FieldOffset(281)]
			public ushort vDebug3;			
			[FieldOffset(283)]
			public ushort vDebug4;			

			[FieldOffset(285)]
			public ushort nDump;						// memory dump counter

			// this is used to mask the output ports
			[FieldOffset(287)]
			public fixed byte _ports[5];				// ={ 0,3,0,0,0 };
			[FieldOffset(292)]
			public fixed byte _inputs[5];				// ports as read

			[FieldOffset(297)]
			public byte RAM_END;						// 'F'

			[FieldOffset(0)]
			public fixed byte ubx[298];
			public void put(int i, byte b){ fixed(byte* p=ubx){ p[i] = b; } }
			public byte get(int i){ fixed(byte* p=ubx){ return p[i]; } }
			public bool port(int i, int b){ fixed(byte* p=_inputs){ return ((p[i]>>b)&1)!=0; } }
			public ushort height(int i){ fixed(ushort* p=_height){ return p[i]; } }	// 0-150
			public ushort rheight(int i){ fixed(ushort* p=adRaw){ return p[i]; } }	// 0-1023
		}
		private EAS eas;
		// The code in the CPU is:
		// usbText("M ");
		// x._inputs[0] = PORTA;
		// x._inputs[1] = PORTB;
		// x._inputs[2] = PORTC;
		// x._inputs[3] = PORTD;
		// x._inputs[4] = PORTE;
		// p1 = &x.RAM_START;
		// p2 = &x.RAM_END;
		// usbWord((int)p2-(int)p1+1);
		// usbText("\r\n");
		// memDump();

		// which returns something like:
		// M 012A
		// 530015028C040000B801B80100000000
		// 90008D00960096008F008C0096009600
		// 90008C009600960090008C0096009600
		// 91008C00960096008F008C0096009600
		// 90008D009600960090008D0096009600
		// 90008C00960096008F008C0096009600
		// 8F008C009600960090008C0096009600
		// 91008D009600960090008C0096009600
		// 90008D00960096008F008B0096009600
		// 90008D00960096008F008C0096009600
		// 90008D009600960090008C0096009600
		// 90008C009600960090008D0096009600
		// 06330000000000000000000000000000
		// 000D0000000000000000000000000000
		// 0000000000000100000002060349000A
		// 30497082000000FAFA00000000FF0101
		// 01010101010101010100FFFFFF0E0307
		// 03070304030000000000000000100108
		// 01000000091940000046
		//
		// Well the length is x12A = 298 as expected and we can see the 'S' and 'F' at the begining and end
		int easlength = (6+2)+18*(32+2)+(20+2);
		int hex(char c){ if(c>='0' && c<='9') return c-'0'; if(c>='A' && c<='F') return c-'A'+10; return 0; }
		int hex(string s, int i){ return (hex(s[i])<<4) + hex(s[i+1]); }
		bool unpack(string s){
			try{
				if(s.Length<easlength) return false;
				if(string.Compare("M 012A\r\n", s.Remove(8))!=0) return false;
				int i=0, j=8;
				while(i<298 && j<s.Length-1){
					if(s[j]=='\r' || s[j]=='\n')
						j++;
					else{
						eas.put(i++, (byte)hex(s, j));
						j+=2;
					}
				}
				byte a = eas.get(297);
				byte b = eas.RAM_END;
				byte[] z = new byte[298];
				for(i=0; i<298; i++)
					z[i] = eas.get(i);
				return i==298;
			}
			catch{
				bugbox.Text += "Threw in unpack()\r\n";
				return false;
			}
		}
		//=============================================================================================================
		// Serial port stuff
		//=============================================================================================================
		SerialPort com=null;
		Thread reader=null;
		private DispatcherTimer timer;
		bool running=false;
		bool first = true;
		string log="";
		string vstring="";

		bool px = false;
		string pxs = "";

		// reader tick
		private void tick(object sender, EventArgs e){
			pollEAS();
//			bugbox.Text = log;
//			log = "";
		}
		// apres run clean up
		public void killthreads(Object sender, EventArgs e){
			running = false;
			if(reader!=null)
				reader.Join();
			reader = null;
		}
		// 'read' thread
		public static void Read(object w1)
		{
			Window1 w = w1 as Window1;
			while(w!=null && w.running && w.com!=null && w.com.IsOpen){
				try{
					string message = w.com.ReadExisting();
					if(!String.IsNullOrEmpty(message)){
						w.log += message;
						if(w.px){
							w.pxs += message;
							if(w.pxs.Length>=w.easlength) w.px = false;
						}
					}
					else
						Thread.Sleep(100);
				}
				catch(TimeoutException){}
				catch{}
			}
		}
		// sent a string to the EAS
		private void Send(string st){
			try{
				if(com!=null && com.IsOpen){
					com.Write(st+"\r");
					log += st+"\r\n";
				}
			}
			catch{
				bugbox.Text += "Threw in Send()\r\n";
			}
		}
		// poll the eas and convert the results
		void pollEAS(){
			try{
				if(com!=null && com.IsOpen){
					int t=1000;			// ten secons in cycles
					if(first){
						px = true;
						pxs = "";
						Send("p");
						first = false;
						while(!pxs.Contains("\n") && t-->0)
							Thread.Sleep(10);
						vstring = pxs.Substring(0,pxs.Length-2);		// remove \r\n
						vstr.Content = vstring;
					}
					else{
						px = true;
						pxs = "";
						Send("d");
						while(px && t-->0) Thread.Sleep(10);
						if(unpack(pxs)){
							logbox.Text = pxs;
							fileLog();
						}
						paint();
					}
				}
			}
			catch{
				bugbox.Text += "Threw in pollEAS()\r\n";
			}
		}
		// copy the values in EAS to the screen
		void valve(Image i, bool open, bool horz){
			string s;
			if(open && horz) s = "openh.png";
			else if(open)	 s = "open.png";
			else if(horz)	 s = "closeh.png";
			else			 s = "close.png";
			i.Source = new BitmapImage(new Uri("/XEAS;component/"+s, UriKind.Relative));
		}
		bool rawdisp = false;		// raw display
		bool subvert = false;
		void paint(){
			try{
				// outputs
//#define xLED1		_PORTA,3
//#define xLED2		_PORTA,4
				valve(frv, eas.port(3,0), false);	//#define xSRF		_PORTD,0
				valve(flv, eas.port(3,1), false);	//#define xSLF		_PORTD,1
				valve(rrv, eas.port(3,2), false);	//#define xSRR		_PORTD,2
				valve(rlv, eas.port(3,3), false);	//#define xSLR		_PORTD,3
				valve(exv, eas.port(3,4), false);	//#define xSEXH		_PORTD,4
				valve(inv, eas.port(3,5), true);	//#define xSINP		_PORTD,5
//#define xRHDATA		_PORTB,0
//#define xRHCLK		_PORTB,1
				valve(comp, eas.port(1,5), false);	//#define xCOMP		_PORTB,5

				// inputs
				up.Content = eas.port(4,0) ? "" : "UP pressed";									//#define xUP		_PORTE,0
				down.Content = eas.port(4,1) ? "" : "DOWN pressed";								//#define xDOWN		_PORTE,1
				subvert = !eas.port(4,2);
				inhibit.Content = subvert ?"INHIBIT pressed" : "";								//#define xINHIBIT	_PORTE,2
				doors.Content = eas.port(2,0) ? "Doors closed" : "Doors open";					//#define xDOORS	_PORTC,0
				overheat.Content = eas.port(3,6) ? "Overheat" : "";								//#define xOVERHEAT	_PORTD,6
				tank.Fill = new SolidColorBrush(eas.port(3,7) ? Colors.LightGreen : Colors.Red);//#define xPRESSURE	_PORTD,7
				brakes.Content = eas.port(1,2) ? "Brakes applied" : "Brakes released";			//#define xBRAKES	_PORTB,2
				handbrake.Content = eas.port(1,3) ? "Handbrake released" : "Handbrake applied";	//#define xROLL		_PORTB,3
				sw2.Content = eas.port(1,4) ? "SW2 released" : "SW2 pressed";					//#define xSW2		_PORTB,4
				// numbers
				rpm.Content = string.Format("{0}", eas.RPM);
				mph.Content = string.Format("{0}", eas.MPH);
				// leg heights
				int i;
				int mv = rawdisp ? 1023 : 150;
				fr_val.Maximum = mv;
				fr_val.Value = i = rawdisp ? eas.rheight(0) : eas.height(0);
				fr_n.Content = string.Format("{0}", i);
				fl_val.Maximum = mv;
				fl_val.Value = i = rawdisp ? eas.rheight(1) : eas.height(1);
				fl_n.Content = string.Format("{0}", i);
				rr_val.Maximum = mv;
				rr_val.Value = i = rawdisp ? eas.rheight(2) : eas.height(2);
				rr_n.Content = string.Format("{0}", i);
				rl_val.Maximum = mv;
				rl_val.Value = i = rawdisp ? eas.rheight(3) : eas.height(3);
				rl_n.Content = string.Format("{0}", i);
		
				target.Content = string.Format("Target {0}", eas.tHeight);

				string[] modes = { "BUMPSTOPS", "ACCESS", "LOW", "NORMAL", "HIGH", "VERYHIGH", "UNKNOWN" };
				smode.Content  = string.Format("Mode  {0}", modes[eas.sMode]);

				string[] state = {  "HOLD",																		// no planned move
									"UP RF",   "UP LF",   "UP RR",   "UP LR",   "UP F",   "UP R",				// up moves
									"DOWN RF", "DOWN LF", "DOWN RR", "DOWN LR", "DOWN F", "DOWN R", "DOWN" };	// down moves
				byte st = (byte)(eas.mState & 0x7f);
				string sst = st>13 ? string.Format("BUG({0})", st) : state[st]; 
				bool settle = (eas.mState & 0x80)!=0;
				tstate.Content = string.Format("State  {0}  {1}", sst, settle ? "(SETTLE)" : "");
				hold.Content = eas.bHold!=0 ? "HOLD" : "RUN";
				compt.Content = string.Format("cTime: {0} {1}", eas.compTimer, eas.compON);

				fru_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				frd_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				rru_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				rrd_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				flu_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				fld_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				rlu_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
				rld_but.Visibility = subvert ? Visibility.Visible : Visibility.Hidden;
			}
			catch{
				bugbox.Text += "Threw in paint()\r\n";
			}
		}

		void toggle(int port, int bit){
			if(subvert && com!=null)
				Send(string.Format("t{0}{1}", port, bit));
		}

		StreamWriter sw=null;
		public Window1(){
			InitializeComponent();
			Large(false);
			// add the ports to the menu
			foreach(string s in SerialPort.GetPortNames()){
				MenuItem m = new MenuItem();
				m.Header = s;
				m.IsCheckable = true;
				m.IsChecked = false;
				m.Click += PortClick;
				mports.Items.Add(m);
			}
			// start a ticker to run the i/o
			timer = new DispatcherTimer();
			timer.Interval = new TimeSpan(0,0,1);
			timer.Tick += tick;
			timer.Start();
			// add a trick to make sure things stop on schedule
			this.Closed += killthreads;
		}
		~Window1(){
			if(com!=null){ com.Close(); com = null; }
		}
		// Menu items
		private void PortClick(object sender, RoutedEventArgs e){
            MenuItem m = e.Source as MenuItem;
			// close old port
			if(com!=null){ com.Close(); com = null; }
			// clean up uld reader
			running = false;
			if(reader!=null){
				reader.Join();		// block until it ends
				reader = null;
			}
			// open the new port
			try{
				// tick the right port
				foreach(MenuItem m1 in mports.Items)
					m1.IsChecked = m==m1;
				// open the new port (600/8/1/n in wineas)
				com = new SerialPort(m.Header.ToString(), 9600, Parity.None, 8, StopBits.One);
				com.ReadTimeout = 500;
				com.WriteTimeout = 500;
				com.Handshake = Handshake.None;
				com.Open();
				bugbox.Text += "Opened Port "+m.Header.ToString()+"\r\n";
				// start a new reader
				running = true;
				reader = new Thread(Read);
				reader.SetApartmentState(ApartmentState.STA);
				reader.Start(this);
				bugbox.Text += "Started reader on Port "+m.Header.ToString()+"\r\n";
			}
			catch(InvalidOperationException){
				bugbox.Text += "InvalidOperationException problem opening Port "+m.Header.ToString()+"\r\n";
			}
			catch(ArgumentOutOfRangeException){
				bugbox.Text += "ArgumentOutOfRangeException problem opening Port "+m.Header.ToString()+"\r\n";
			}
			catch(ArgumentException){
				bugbox.Text += "ArgumentException problem opening Port "+m.Header.ToString()+"\r\n";
			}
			catch(IOException ex){
				bugbox.Text += "IOException " + ex.Message + " problem opening Port "+m.Header.ToString()+"\r\n";
			}
			catch(UnauthorizedAccessException){
				bugbox.Text += "UnauthorizedAccessException problem opening Port "+m.Header.ToString()+"\r\n";
			}
			catch{
				bugbox.Text += "Problem opening Port "+m.Header.ToString()+"\r\n";
			}
        }

		private void Ping(object sender, RoutedEventArgs e){
			first = true;
		}
		private void Poll(object sender, RoutedEventArgs e){
			pollEAS();
		}
		bool wasLarge=false;
		private void Large(bool big){
			wasLarge = big;
			if(big){ Height = 820; Width=800;}
			else   { Height = 465; Width=553; }
		}
		private void Large(object sender, RoutedEventArgs e){
			Large(!wasLarge);
		}
		private void Exit(object sender, RoutedEventArgs e){
			Close();
		}
		// buttons on legs (active only if in subvert)
		// legs FR=0, FL=1, RR=2, RL=3
		const int FR=0, FL=1, RR=2, RL=3;	// leg
		const int STOP=0, UP=1, DOWN=2;		// up/down
		void vset(int port, int bit, bool v){
			if(eas.port(port, bit)!=v)
				toggle(port, bit);
		}
		void ud(int leg, int uds){
			bool vs = uds!=STOP;
			switch(leg){
			case FR:	vset(3, 0, vs);		break;
			case FL:	vset(3, 1, vs);		break;
			case RR:	vset(3, 2, vs);		break;
			case RL:	vset(3, 3, vs);		break;
			}
			switch(uds){
			case STOP:	vset(3, 5, false);	vset(3,4, false); break;	// exhaust ; inlet
			case UP:	vset(3, 5, false);	vset(3,4, true); break;
			case DOWN:	vset(3, 5, true);	vset(3,4, false); break;
			}
		}
		// front right
		private void fru_press(object sender, MouseButtonEventArgs e)  { ud(FR, UP); }
		private void fru_release(object sender, MouseButtonEventArgs e){ ud(FR, STOP); }
		private void frd_press(object sender, MouseButtonEventArgs e)  { ud(FR, DOWN); }
		private void frd_release(object sender, MouseButtonEventArgs e){ ud(FR, STOP); }
		// rear right
		private void rru_press(object sender, MouseButtonEventArgs e)  { ud(RR, UP); }
		private void rru_release(object sender, MouseButtonEventArgs e){ ud(RR, STOP); }
		private void rrd_press(object sender, MouseButtonEventArgs e)  { ud(RR, DOWN); }
		private void rrd_release(object sender, MouseButtonEventArgs e){ ud(RR, STOP); }
		// front left
		private void flu_press(object sender, MouseButtonEventArgs e)  { ud(FL, UP); }
		private void flu_release(object sender, MouseButtonEventArgs e){ ud(FL, STOP); }
		private void fld_press(object sender, MouseButtonEventArgs e)  { ud(FL, DOWN); }
		private void fld_release(object sender, MouseButtonEventArgs e){ ud(FL, STOP); }
		// rear left
		private void rlu_press(object sender, MouseButtonEventArgs e)  { ud(RL, UP); }
		private void rlu_release(object sender, MouseButtonEventArgs e){ ud(RL, STOP); }
		private void rld_press(object sender, MouseButtonEventArgs e)  { ud(RL, DOWN); }
		private void rld_release(object sender, MouseButtonEventArgs e){ ud(RL, STOP); }
		// valves (active only if mode==CONTROL)
		private void frv_press(object sender, MouseButtonEventArgs e)  { toggle(3,0); }		// right front
		private void frv_release(object sender, MouseButtonEventArgs e){}
		private void flv_press(object sender, MouseButtonEventArgs e)  { toggle(3,1); }		// left front
		private void flv_release(object sender, MouseButtonEventArgs e){}
		private void rrv_press(object sender, MouseButtonEventArgs e)  { toggle(3,2); }		// right rear
		private void rrv_release(object sender, MouseButtonEventArgs e){}
		private void rlv_press(object sender, MouseButtonEventArgs e)  { toggle(3,3); }		// left rear
		private void rlv_release(object sender, MouseButtonEventArgs e){}
		private void inv_press(object sender, MouseButtonEventArgs e)  { toggle(3,5); }		// exhaust
		private void inv_release(object sender, MouseButtonEventArgs e){}
		private void exv_press(object sender, MouseButtonEventArgs e)  { toggle(3,4); }		// inlet
		private void exv_release(object sender, MouseButtonEventArgs e){}
		private void comp_press(object sender, MouseButtonEventArgs e) { toggle(1,5); }		// compresser
		private void comp_release(object sender, MouseButtonEventArgs e){}

		// 'raw' display checkbox
		private void raw(object sender, RoutedEventArgs e){
			rawdisp = !rawdisp;
		}

		void fileLog()
		{
			try{
				if(sw==null){
					string filePath = new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath;
					string logpath = System.IO.Path.GetDirectoryName(filePath)+"\\XAS.log";
					sw = new StreamWriter(logpath);
					sw.Write("New: {0} {1}\r\n", vstr.Content, DateTime.Now.ToShortDateString());
				}
				sw.Write("{0} ", DateTime.Now.ToShortTimeString());
				sw.Write("SRF {0} ", eas.port(3,0)?1:0);
				sw.Write("SLF {0} ", eas.port(3,1)?1:0);
				sw.Write("SRR {0} ", eas.port(3,2)?1:0);
				sw.Write("SRL {0} ", eas.port(3,3)?1:0);
				sw.Write("SEX {0} ", eas.port(3,4)?1:0);
				sw.Write("SIN {0} ", eas.port(3,5)?1:0);
				sw.Write("CMP {0} ", eas.port(1,5)?1:0);

				// inputs
				sw.Write(eas.port(4,0) ? "   " : "UP ");
				sw.Write(eas.port(4,1) ? "     " : "DOWN ");
				// subvert = !eas.port(4,2);
				sw.Write(eas.port(4,2) ? "    ":"INH");
				sw.Write(eas.port(2,0) ? "     " : "DOOR ");
				sw.Write(eas.port(3,6) ? "OH! " : "    ");
				sw.Write(eas.port(3,7) ? "     ": "PBAD ");
				sw.Write(eas.port(1,2) ? "BRAKES " : "       ");
				sw.Write(eas.port(1,3) ? "HP " : "   ");
				// sw2.Content = eas.port(1,4) ? "SW2 released" : "SW2 pressed";					//#define xSW2		_PORTB,4
				// numbers
				sw.Write("RPM {0:D4 }", eas.RPM);
				sw.Write("MPH {0:D3 }", eas.MPH);
				// leg heights
				//	int i;
				//	int mv = rawdisp ? 1023 : 150;
				//	fr_val.Maximum = mv;
				sw.Write("LF: {0:D3},{1:D3} ", eas.rheight(0), eas.height(0));
				sw.Write("RF: {0:D3},{1:D3} ", eas.rheight(1), eas.height(1));
				sw.Write("LR: {0:D3},{1:D3} ", eas.rheight(2), eas.height(2));
				sw.Write("RR: {0:D3},{1:D3} ", eas.rheight(3), eas.height(3));

				sw.Write("TG {0} ", eas.tHeight);

				string[] modes = { "BUMP", "ACCS", "LOW ", "NORM", "HIGH", "VHGH", "UNKN" };
				sw.Write("Mode {0} ", modes[eas.sMode]);

				string[] state = {  "HOLD",														// no planned move
									"U RF", "U LF", "U RR", "U LR", "UP F", "UP R",				// up moves
									"D RF", "D LF", "D RR", "D LR", "DN F", "DN R", "DOWN" };	// down moves
				byte st = (byte)(eas.mState & 0x7f);
				string sst = st>13 ? string.Format("B({0})", st) : state[st]; 
				bool settle = (eas.mState & 0x80)!=0;
				sw.Write("STS {0} {1} ", sst, settle ? "(SE)" : "    ");
				sw.Write(eas.bHold!=0 ? "HOLD" : "RUN ");
				sw.Write("CompT: {0:D3} ", eas.compTimer);
				sw.Write("\r\n");
				sw.Flush();
			}
			catch{
				bugbox.Text += "Threw in fileLog()\r\n";
			}
		}
	}
}
